import Interpolator from "../../math/interpolator";

/**
 * @constructor
 * @param {MdxSd} sd
 * @param {number} start
 * @param {number} end
 * @param {Array<MdxParserTrack>} keyframes
 * @param {boolean} isGlobalSequence
 */
function MdxSdSequence(sd, start, end, keyframes, isGlobalSequence) {
    var defval = sd.defval;

    this.sd = sd;
    this.start = start;
    this.end = end;
    this.keyframes = [];

    // When using a global sequence, where the first key is outside of the sequence's length, it becomes its constant value.
    // When having one key in the sequence's range, and one key outside of it, results seem to be non-deterministic.
    // Sometimes the second key is used too, sometimes not.
    // It also differs depending where the model is viewed - the WE previewer, the WE itself, or the game.
    // All three show different results, none of them make sense.
    // Therefore, only handle the case where the first key is outside.
    // This fixes problems spread over many models, e.g. HeroMountainKing (compare in WE and in Magos).
    if (isGlobalSequence && keyframes[0].frame > end) {
        this.keyframes.push(keyframes[0]);
    }

    // Go over the keyframes, and add all of the ones that are in this sequence (start <= frame <= end).
    for (var i = 0, l = keyframes.length; i < l; i++) {
        var keyframe = keyframes[i],
            frame = keyframe.frame;

        if (frame >= start && frame <= end) {
            this.keyframes.push(keyframe);
        }
    }

    switch (this.keyframes.length) {
        // If there are no keys, use the default value directly.
        case 0:
            this.constant = true;
            this.value = defval;
            break;

        // If there's only one key, use it directly.
        case 1:
            this.constant = true;
            this.value = this.keyframes[0].value;
            break;

        default:
            // If all of the values in this sequence are the same, might as well make it constant.
            var constant = true,
                firstValue = this.keyframes[0].value;

            for (var i = 1, l = this.keyframes.length; i < l; i++) {
                var keyframe = this.keyframes[i],
                    value = keyframe.value;

                if (value.length > 0) {
                    for (var j = 0, k = value.length; j < k; j++) {
                        if (firstValue[j] !== value[j]) {
                            constant = false;
                            break;
                        }
                    }
                } else {
                    if (value !== firstValue) {
                        constant = false;
                        break;
                    }
                }
            }

            if (constant) {
                this.constant = true;
                this.value = firstValue;
            } else {
                this.constant = false;

                // If there is no opening keyframe for this sequence, inject one with the default value.
                if (this.keyframes[0].frame !== start) {
                    this.keyframes.splice(0, 0, { frame: start, value: defval, inTan: defval, outTan: defval });
                }

                // If there is no closing keyframe for this sequence, inject one with the default value.
                if (this.keyframes[this.keyframes.length - 1].frame !== end) {
                    this.keyframes.splice(this.keyframes.length, 0, { frame: end, value: this.keyframes[0].value, inTan: this.keyframes[0].outTan, outTan: this.keyframes[0].inTan });
                }
            }

            break;
    }

    this.keyframeInterval = new Uint32Array(2);
}

MdxSdSequence.prototype = {
    cacheValue(frame) {
        if (this.constant) {
            this.cachedIndex = 0;

            return this.value;
        } else {
            var keyframes = this.keyframes,
                l = keyframes.length;

            if (frame <= this.start) {
                this.cachedIndex = 0;

                return keyframes[0].value;
            } else if (frame >= this.end) {
                this.cachedIndex = l - 1;

                return keyframes[l - 1].value;
            } else {
                for (var i = 1; i < l; i++) {
                    var keyframe = keyframes[i];

                    if (keyframe.frame > frame) {
                        var lastKeyframe = keyframes[i - 1];

                        var t = Math.clamp((frame - lastKeyframe.frame) / (keyframe.frame - lastKeyframe.frame), 0, 1);

                        this.cachedIndex = i;

                        return Interpolator.interpolate(lastKeyframe.value, lastKeyframe.outTan, keyframe.inTan, keyframe.value, t, this.sd.interpolationType);
                    }
                }
            }
        }
    },

    getValue1V2(out, frame) {
        let value = this.cacheValue(frame);

        out[0] = value[0];

        return this.cachedIndex;
    },

    getValue3V2(out, frame) {
        let value = this.cacheValue(frame);

        out[0] = value[0];
        out[1] = value[1];
        out[2] = value[2];

        return this.cachedIndex;
    },

    getValue4V2(out, frame) {
        let value = this.cacheValue(frame);

        out[0] = value[0];
        out[1] = value[1];
        out[2] = value[2];
        out[3] = value[3];

        return this.cachedIndex;
    },

    getValue1(out, frame) {
        let value,
            index;

        if (this.constant) {
            value = this.value;
            index = 0;
        } else {
            var keyframes = this.keyframes,
                l = keyframes.length;

            if (frame <= this.start) {
                value = keyframes[0].value;
                index = 0;
            } else if (frame >= this.end) {
                value = keyframes[l - 1].value;
                index = l - 1;
            } else {
                for (var i = 1; i < l; i++) {
                    var keyframe = keyframes[i];

                    if (keyframe.frame > frame) {
                        var lastKeyframe = keyframes[i - 1];

                        var t = Math.clamp((frame - lastKeyframe.frame) / (keyframe.frame - lastKeyframe.frame), 0, 1);

                        value = Interpolator.vector(lastKeyframe.value, lastKeyframe.outTan, keyframe.inTan, keyframe.value, t, this.sd.interpolationType);
                        index = i;

                        break;
                    }
                }
            }
        }

        out[0] = value[0];

        return index;
    },

    getValue3(out, frame) {
        let value,
            index;

        if (this.constant) {
            value = this.value;
            index = 0;
        } else {
            var keyframes = this.keyframes,
                l = keyframes.length;

            if (frame <= this.start) {
                value = keyframes[0].value;
                index = 0;
            } else if (frame >= this.end) {
                value = keyframes[l - 1].value;
                index = l - 1;
            } else {
                for (var i = 1; i < l; i++) {
                    var keyframe = keyframes[i];

                    if (keyframe.frame > frame) {
                        var lastKeyframe = keyframes[i - 1];

                        var t = Math.clamp((frame - lastKeyframe.frame) / (keyframe.frame - lastKeyframe.frame), 0, 1);

                        value = Interpolator.vector(lastKeyframe.value, lastKeyframe.outTan, keyframe.inTan, keyframe.value, t, this.sd.interpolationType);
                        index = i;

                        break;
                    }
                }
            }
        }

        out[0] = value[0];
        out[1] = value[1];
        out[2] = value[2];

        return index;
    },

    getValue4(out, frame) {
        let value,
            index;

        if (this.constant) {
            value = this.value;
            index = 0;
        } else {
            var keyframes = this.keyframes,
                l = keyframes.length;

            if (frame <= this.start) {
                value = keyframes[0].value;
                index = 0;
            } else if (frame >= this.end) {
                value = keyframes[l - 1].value;
                index = l - 1;
            } else {
                for (var i = 1; i < l; i++) {
                    var keyframe = keyframes[i];

                    if (keyframe.frame > frame) {
                        var lastKeyframe = keyframes[i - 1];

                        var t = Math.clamp((frame - lastKeyframe.frame) / (keyframe.frame - lastKeyframe.frame), 0, 1);

                        value = Interpolator.quaternion(lastKeyframe.value, lastKeyframe.outTan, keyframe.inTan, keyframe.value, t, this.sd.interpolationType);
                        index = i;

                        break;
                    }
                }
            }
        }

        out[0] = value[0];
        out[1] = value[1];
        out[2] = value[2];
        out[3] = value[3];

        return index;
    },

    getValueUnsafe(frame) {
        if (this.constant) {
            return this.value;
        } else {
            var keyframes = this.keyframes,
                l = keyframes.length;

            if (frame <= this.start) {
                return keyframes[0].value;
            } else if (frame >= this.end) {
                return keyframes[l - 1].value;
            } else {
                for (var i = 1; i < l; i++) {
                    var keyframe = keyframes[i];

                    if (keyframe.frame > frame) {
                        var lastKeyframe = keyframes[i - 1];

                        var t = Math.clamp((frame - lastKeyframe.frame) / (keyframe.frame - lastKeyframe.frame), 0, 1);

                        return Interpolator.interpolate(lastKeyframe.value, lastKeyframe.outTan, keyframe.inTan, keyframe.value, t, this.sd.interpolationType);
                    }
                }
            }
        }
    },

    getKeyframe(frame) {
        if (this.constant) {
            return 0;
        } else {
            var keyframes = this.keyframes,
                l = keyframes.length;

            if (frame <= this.start) {
                return 0
            } else if (frame >= this.end) {
                return l - 1;
            } else {
                for (var i = 1; i < l; i++) {
                    var keyframe = keyframes[i];

                    if (keyframe.frame > frame) {
                        return i;
                    }
                }
            }
        }
    }
};

var forcedInterpMap = {
    KLAV: 0,
    KATV: 0,
    KPEV: 0,
    KP2V: 0,
    KRVS: 0
};

/**
 * @constructor
 * @param {MdxModel} model
 * @param {MdxParserSd} sd
 */
function MdxSd(model, sd) {
    var globalSequenceId = sd.globalSequenceId,
        globalSequences = model.globalSequences,
        tracks = sd.tracks,
        forcedInterp = forcedInterpMap[sd.tag];

    this.tag = sd.tag;
    this.model = model;
    this.keyframes = tracks;
    this.defval = sd.defval;

    // Allow to force an interpolation type.
    // The game seems to do this with visibility tracks, where the type is forced to None.
    // It came up as a bug report by a user who used the wrong interpolation type.
    this.interpolationType = forcedInterp !== undefined ? forcedInterp : sd.interpolationType;

    if (globalSequenceId !== -1 && globalSequences) {
        this.globalSequence = new MdxSdSequence(this, 0, globalSequences[globalSequenceId].value, tracks, true);
    } else {
        var sequences = model.sequences;

        this.sequences = [];

        for (var i = 0, l = sequences.length; i < l; i++) {
            var interval = sequences[i].interval;

            this.sequences[i] = new MdxSdSequence(this, interval[0], interval[1], tracks, false);
        }
    }
}

MdxSd.prototype = {
    getValue1(out, instance) {
        if (this.globalSequence) {
            var globalSequence = this.globalSequence;

            return globalSequence.getValue1(out, instance.counter % globalSequence.end);
        } else if (instance.sequence !== -1) {
            return this.sequences[instance.sequence].getValue1(out, instance.frame);
        } else {
            let defval = this.defval;

            out[0] = defval[0];

            return 0;
        }
    },

    getValue3(out, instance) {
        if (this.globalSequence) {
            var globalSequence = this.globalSequence;

            return globalSequence.getValue3(out, instance.counter % globalSequence.end);
        } else if (instance.sequence !== -1) {
            return this.sequences[instance.sequence].getValue3(out, instance.frame);
        } else {
            let defval = this.defval;

            out[0] = defval[0];
            out[1] = defval[1];
            out[2] = defval[2];

            return 0;
        }
    },

    getValue4(out, instance) {
        if (this.globalSequence) {
            var globalSequence = this.globalSequence;

            return globalSequence.getValue4(out, instance.counter % globalSequence.end);
        } else if (instance.sequence !== -1) {
            return this.sequences[instance.sequence].getValue4(out, instance.frame);
        } else {
            let defval = this.defval;

            out[0] = defval[0];
            out[1] = defval[1];
            out[2] = defval[2];
            out[3] = defval[3];

            return 0;
        }
    },

    getValueUnsafe(instance) {
        if (this.globalSequence) {
            var globalSequence = this.globalSequence;

            return globalSequence.getValueUnsafe(instance.counter % globalSequence.end);
        } else if (instance.sequence !== -1) {
            return this.sequences[instance.sequence].getValueUnsafe(instance.frame);
        } else {
            return this.defval;
        }
    },

    getKeyframe(instance) {
        if (this.globalSequence) {
            var globalSequence = this.globalSequence;

            return globalSequence.getKeyframe(instance.counter % globalSequence.end);
        } else if (instance.sequence !== -1) {
            return this.sequences[instance.sequence].getKeyframe(instance.frame);
        } else {
            return 0;
        }
    },

    isVariant(sequence) {
        if (this.globalSequence) {
            return !this.globalSequence.constant;
        } else {
            return !this.sequences[sequence].constant;
        }
    },

    getValues() {
        if (this.globalSequence) {
            var values = [],
                keyframes = this.globalSequence.keyframes;

            for (var i = 0, l = keyframes.length; i < l; i++) {
                values[i] = keyframes[i].value;
            }

            return values;
        } else {
            console.warn("[MdxSD::getValues] Called on an SD that doesn't use a global sequence")
            return [];
        }
    }
};

/**
 * @constructor
 * @param {MdxModel} model
 * @param {MdxParserSdContainer} container
 */
function MdxSdContainer(model, container) {
    let sd = {},
		elements = container.elements;

	for (let i = 0, l = elements.length; i < l; i++) {
		let element = elements[i];

		sd[element.tag] = new MdxSd(model, element);
	}

    this.sd = sd;
}

MdxSdContainer.prototype = {
    getValues(tag) {
        var sd = this.sd[tag];

        if (sd) {
            return sd.getValues();
        }

        return [];
    },

    getValue1V2(out, tag, instance, defval) {
        var sd = this.sd[tag];

        if (sd) {
            return sd.getValue1(out, instance);
        }

        out[0] = defval;

        return 0;
    },

    getValue3V2(out, tag, instance, defval) {
        var sd = this.sd[tag];

        if (sd) {
            return sd.getValue3(out, instance);
        }

        out[0] = defval[0];
        out[1] = defval[1];
        out[2] = defval[2];

        return 0;
    },

    getValue4V2(out, tag, instance, defval) {
        var sd = this.sd[tag];

        if (sd) {
            return sd.getValue4(out, instance);
        }

        out[0] = defval[0];
        out[1] = defval[1];
        out[2] = defval[2];
        out[3] = defval[3];

        return 0;
    },

    getValueUnsafe(tag, instance, defval) {
        var sd = this.sd[tag];

        if (sd) {
            return sd.getValueUnsafe(instance);
        }

        return defval;
    },

    getValue(tag, instance, defval) {
        return this.getValueUnsafe(tag, instance, defval);
    },

    getValue2(out, tag, instance, defval) {
        let unsafeHeap = this.getValueUnsafe(tag, instance, defval);

        out[0] = unsafeHeap[0];
        out[1] = unsafeHeap[1];

        return out;
    },

    getValue3(out, tag, instance, defval) {
        let unsafeHeap = this.getValueUnsafe(tag, instance, defval);

        out[0] = unsafeHeap[0];
        out[1] = unsafeHeap[1];
        out[2] = unsafeHeap[2];

        return out;
    },

    getValue4(out, tag, instance, defval) {
        let unsafeHeap = this.getValueUnsafe(tag, instance, defval);

        out[0] = unsafeHeap[0];
        out[1] = unsafeHeap[1];
        out[2] = unsafeHeap[2];
        out[3] = unsafeHeap[3];

        return out;
    },

    getKeyframe(tag, instance) {
        var sd = this.sd[tag];

        if (sd) {
            return sd.getKeyframe(instance);
        }

        return 0;
    },

    isVariant(tag, sequence) {
        var sd = this.sd[tag];

        if (sd) {
            return sd.isVariant(sequence);
        }

        return false;
    }
};

export default MdxSdContainer;
